import inps
import pandas as pd
import numpy as np
from sklearn.naive_bayes import BernoulliNB
from sklearn.neural_network import MLPRegressor
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from xgboost import XGBRegressor, XGBClassifier

np_sample = pd.read_parquet("./data/ED/nonprobEd.parquet")
p_sample = pd.read_parquet("./data/ED/probEd.parquet")
population_size = 40884010

psa_covariates = ["Altura", "Peso", "Estudios", "Frecuencia_acf", "IMC_rec"]
calibration_covariates = ["Sexo", "Edad_rec"]

levels = {
	"Estudios": ('Educación general básica u obligatoria (primaria, ESO)', 'Grado medio o Formación Profesional I (FP1)', 'Bachillerato, Grado Superior o Formación Profesional II (FP2)', 'Estudios Universitarios (Grado, Diplomatura o Licenciatura)', 'Estudios de Postgrado (Máster, Doctorado)'),
	"Frecuencia_acf": ('Rara vez', '1 vez por semana', '2-3 veces por semana', 'Casi todos los días'),
	"IMC_rec": ('Infrapeso', 'Normopeso', 'Sobrepeso', 'Obesidad')
}

for sample in np_sample, p_sample:
	for column, ordered_values in levels.items():
		sample[column] = sample[column].cat.rename_categories(lambda code: ordered_values.index(code)).astype("Int64")

psa_weights = inps.psa_weights(np_sample, p_sample, population_size, covariates = psa_covariates)
kw_weights = inps.kw_weights(np_sample, p_sample, population_size, covariates = psa_covariates)

psa_weights2 = inps.psa_weights(np_sample, p_sample, population_size, covariates = psa_covariates, model = inps.make_preprocess_estimator(BernoulliNB()))
kw_weights2 = inps.kw_weights(np_sample, p_sample, population_size, covariates = psa_covariates, model = XGBClassifier())

target_var = "D4"
naive_estimation = inps.estimation(np_sample[target_var])
psa_estimation = inps.estimation(np_sample[target_var], psa_weights["np"])
kw_estimation = inps.estimation(np_sample[target_var], kw_weights)

print(naive_estimation, psa_estimation, kw_estimation)

population_totals = pd.Series({'Hombre': 19972404, 'Mujer': 20911606, '18-29': 7406234, '30-44': 9691504, '45+': 23786272})

np_sample["Edad_rec"] = np_sample["Edad_rec"].cat.rename_categories({'1': '18-29', '2': '30-44', '3': '45+'})
np_sample = pd.get_dummies(np_sample, prefix = '', prefix_sep = '', columns = calibration_covariates)

np_sample["weights"] = psa_weights["np"]
psa_calibration_weights = inps.calibration_weights(np_sample, population_totals, weights_column = "weights")
psa_calibration_estimation = inps.estimation(np_sample[target_var], psa_calibration_weights)

print(psa_calibration_estimation)

np_sample = pd.read_parquet("./data/covid/nonprobCovid.parquet")
p_sample = pd.read_parquet("./data/covid/probCovid.parquet")

covariates = ['age', 'gender', 'region', 'household_size', 'employment_status']

p_covariates_names = ['EDAD', 'SEXO', 'CCAA', 'NUMHOGAR', 'SITLAB']
p_sample.rename(columns = dict(zip(p_covariates_names, covariates)), inplace = True)

p_sample['gender'] = p_sample['gender'].replace({1: "Male", 2: "Female"})

regions = ('', 'Andalucía', 'Aragón', 'Principado de Asturias', 'Islas Baleares', 'Islas Canarias', 'Cantabria', 'Castilla-La Mancha', 'Castilla y León', 'Cataluña', 'Comunidad Valenciana', 'Extremadura', 'Galicia', 'Comunidad de Madrid', 'Región de Murcia', 'Comunidad Foral de Navarra', 'País Vasco', 'La Rioja', 'Ceuta', 'Melilla')
p_sample['region'] = p_sample['region'].map(lambda code: regions[code], na_action = 'ignore')

np_sample['household_size'] = pd.to_numeric(np_sample['household_size'].replace({'8 or more': 8}), errors = 'coerce')
p_sample['household_size'] = p_sample['household_size'].replace(99, np.nan)
p_sample.loc[p_sample['household_size'] > 8, 'household_size'] = 8

np_sample['employment_status'] = np_sample['employment_status'].replace({
	'Full time employment': 'Employed', 'Part time employment': 'Employed', 'Full time student': 'Student', 'Retired': 'Unemployed', 'Not working': 'Other'
})
p_status = ('', 'Employed', 'Unemployed', 'Unemployed', 'Unemployed', 'Unemployed', 'Student', 'Other', 'Other', np.nan)
p_sample['employment_status'] = p_sample['employment_status'].map(lambda code: p_status[code], na_action = 'ignore')

for categorical in ('gender', 'region', 'employment_status'):
	for sample in np_sample, p_sample:
		sample[categorical] = sample[categorical].astype('category')

categories = ('Not at all', 'Rarely', 'Sometimes', 'Frequently', 'Always')
np_sample['target_num'] = np_sample['i12_health_7'].map(lambda code: categories.index(code), na_action = 'ignore') + 1

matching_values = inps.matching_values(np_sample, p_sample, "target_num", covariates = covariates, model = inps.make_preprocess_estimator(MLPRegressor()))
training_values = inps.training_values(np_sample, p_sample, "target_num", covariates = covariates, psa_model = XGBClassifier(enable_categorical = True, tree_method = "hist"), matching_model = XGBRegressor(enable_categorical = True, tree_method = "hist"))

naive_estimation = inps.estimation(np_sample['target_num'])
matching_estimation = inps.estimation(matching_values['p'])
training_estimation = inps.estimation(training_values['p'])

dr_estimation = inps.doubly_robust_estimation(np_sample, p_sample, "target_num", covariates = covariates, matching_model = inps.make_preprocess_estimator(KNeighborsRegressor()))

print(naive_estimation, matching_estimation, training_estimation, dr_estimation)

np_sample['target_cat'] = np_sample['target_num'] > 3
training_values = inps.training_values(np_sample, p_sample, "target_cat", True, covariates = covariates)

naive_estimation = inps.estimation(np_sample['target_cat'])
training_estimation = inps.estimation(training_values['p'])

print(naive_estimation, training_estimation)

np_sample = pd.read_parquet("./data/health/nonprobHealth.parquet")
population = pd.read_parquet("./data/health/censusHealth.parquet")
covariates = ["sexo", "b2", "dapae_memo", "PERFILPROFTITUL"]
target_var = "CRONICIDAD"
target_category = "Sí"

np_sample = np_sample[np_sample[target_var].notna()]

imputed_values = inps.training_values(np_sample, population, target_var, target_category, covariates = covariates)
imputed_values2 = inps.matching_values(np_sample, population, target_var, target_category, covariates = covariates, model = inps.make_preprocess_estimator(KNeighborsClassifier()))

naive_estimation = inps.estimation(np_sample[target_var] == target_category)
model_estimation = inps.estimation(imputed_values["p"])
custom_model_estimation = inps.estimation(imputed_values2["p"])

print(naive_estimation, model_estimation, custom_model_estimation)



"""Python package for statistical inference from non-probability samples"""

__version__ = "1.16"

import numpy as np
import pandas as pd
import sklearn
import pandas.api.types as types
from sklearn.compose import ColumnTransformer
from sklearn.compose import make_column_selector as column_selector
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import RobustScaler, OneHotEncoder
from sklearn.model_selection import StratifiedKFold
from sklearn.linear_model import RidgeCV, LogisticRegressionCV
from sklearn.ensemble import HistGradientBoostingRegressor, HistGradientBoostingClassifier
from scipy.stats import bootstrap

sklearn.set_config(enable_metadata_routing = True)

def default_preprocess(**kwargs):
	return ColumnTransformer([
		("numeric", Pipeline([
			("normalizer", RobustScaler()),
			("imputer", SimpleImputer(strategy = 'median', add_indicator = True, copy = False))
		]), column_selector(dtype_include = 'number')),
		("categorical",
			OneHotEncoder(drop = 'if_binary', min_frequency = .05, handle_unknown = 'infrequent_if_exist'),
		column_selector(dtype_exclude = 'number'))
	], **kwargs)

def make_preprocess_estimator(base_estimator, **kwargs):
	if hasattr(base_estimator, "set_fit_request"):
		base_estimator = base_estimator.set_fit_request(sample_weight = True)
	if hasattr(base_estimator, "set_score_request"):
		base_estimator = base_estimator.set_score_request(sample_weight = True)
	
	return Pipeline([
		("preprocess", default_preprocess(**kwargs)),
		("estimator", base_estimator)
	])

def logistic_classifier(**kwargs):
	return make_preprocess_estimator(LogisticRegressionCV(cv = StratifiedKFold(shuffle = True, random_state = 0), scoring = 'neg_log_loss', max_iter = 1000, **kwargs))

def linear_regressor(**kwargs):
	return make_preprocess_estimator(RidgeCV(**kwargs))

valid_dtypes = (types.is_numeric_dtype, lambda dtype: isinstance(dtype, types.CategoricalDtype), types.is_datetime64_any_dtype, types.is_timedelta64_dtype)

def set_categories(data):
	invalid_columns = [column for column, dtype in data.dtypes.items() if not any(is_dtype(dtype) for is_dtype in valid_dtypes)]
	data[invalid_columns] = data[invalid_columns].astype("category", copy = False)

def boosting_classifier(**kwargs):
	my_model = HistGradientBoostingClassifier(categorical_features = 'from_dtype', random_state = 0, early_stopping = True, max_iter = 1000, **kwargs)
	return my_model.set_fit_request(sample_weight = True).set_score_request(sample_weight = True)

def boosting_regressor(**kwargs):
	my_model = HistGradientBoostingRegressor(categorical_features = 'from_dtype', random_state = 0, early_stopping = True, max_iter = 1000, **kwargs)
	return my_model.set_fit_request(sample_weight = True).set_score_request(sample_weight = True)

def calibration_weights(sample, population_totals, weights_column = None, population_size = None, max_steps = 1000, tolerance = 1e-6):
	if weights_column is not None:
		valid = sample[weights_column].notna()
		d = sample.loc[valid, weights_column]
		X = sample.loc[valid, population_totals.index].to_numpy(dtype = 'float64')
		X = X * d.to_numpy().reshape(-1, 1)
	elif population_size is not None:
		X = sample[population_totals.index].to_numpy(dtype = 'float64')
		d = population_size / X.shape[0]
		X = X * d
	else:
		raise ValueError("weights_column or population_size must be set")
	
	T = population_totals.to_numpy()
	L = np.zeros(X.shape[1])
	w = np.ones(X.shape[0])
	success = False
	
	for step in range(max_steps):
		L += np.dot(np.linalg.pinv(np.dot(np.multiply(X.T, w), X)), (T - np.dot(X.T, w)))
		w = np.exp(np.dot(X, L))
		
		loss = np.max(np.abs((np.dot(X.T, w) - T) / T))
		if loss < tolerance:
			success = True
			break
	
	if not success: raise Exception("Calibration did not converge")
	return w * d

def propensities(np_sample, p_sample, weights_column = None, covariates = None, model = None):
	np_size = np_sample.shape[0]
	p_size = p_sample.shape[0]
	weights = np.ones(p_size) if weights_column is None else p_sample[weights_column]
	
	if covariates is not None:
		np_sample = np_sample.loc[:, covariates]
		p_sample = p_sample.loc[:, covariates]
	
	if model is None: model = logistic_classifier()
	
	X = pd.concat((np_sample, p_sample), ignore_index = True, join = "inner", copy = False)
	set_categories(X)
	y = np.concatenate((np.ones(np_size, dtype = bool), np.zeros(p_size, dtype = bool)))
	sample_weight = np.concatenate((np.repeat(np.sum(weights) / np_size, np_size), weights))
	sample_weight /= np.mean(sample_weight)
	model.fit(X, y, sample_weight = sample_weight)
	return model.predict_proba(X)[:, tuple(model.classes_).index(True)]

def psa_weights(np_sample, p_sample, population_size = None, weights_column = None, covariates = None, model = None):
	np_size = np_sample.shape[0]
	my_propensities = propensities(np_sample, p_sample, weights_column, covariates, model)
	optimal_weights = (1 - my_propensities) / my_propensities
	
	if population_size:
		optimal_weights *= population_size / np.sum(optimal_weights[:np_size])
	
	return {"np": optimal_weights[:np_size], "p": optimal_weights[np_size:]}

def matching_values(np_sample, p_sample, target_column, target_category = None, covariates = None, model = None, training_weight = None):
	y = np_sample[target_column]
	if y.isna().any(): raise ValueError("Missing values in target column")
	is_numerical = types.is_numeric_dtype(y) and target_category is None
	if target_category is not None: y = y == target_category
	if model is None: model = linear_regressor() if is_numerical else logistic_classifier()
	
	if covariates is None: covariates = np_sample.columns.intersection(p_sample.columns)
	np_sample = np_sample.loc[:, covariates]
	p_sample = p_sample.loc[:, covariates]
	set_categories(np_sample)
	set_categories(p_sample)
	
	if training_weight is None:
		model.fit(np_sample, y)
	else:
		training_weight = training_weight / np.mean(training_weight)
		model.fit(np_sample, y, sample_weight = training_weight)
	
	def predict(X):
		if is_numerical:
			return model.predict(X)
		else:
			probs = model.predict_proba(X)
			if target_category is None:
				return {"categories": model.classes_, "probs": probs}
			else:
				return probs[:, tuple(model.classes_).index(True)]
	
	return {"p": predict(p_sample), "np": predict(np_sample)}

def doubly_robust_estimation(np_sample, p_sample, target_column, target_category = None, weights_column = None, covariates = None, psa_model = None, matching_model = None):
	if not types.is_numeric_dtype(np_sample[target_column]) and target_category is None:
		raise ValueError("target_category must be set when the target variable is categorical.")
	
	if weights_column is None:
		weights = None
	else:
		weights = p_sample[weights_column]
	
	imputed_weights = psa_weights(np_sample, p_sample, None, weights_column, covariates, psa_model)["np"]
	imputed_values = matching_values(np_sample, p_sample, target_column, target_category, covariates, matching_model)
	original_values = np_sample[target_column] if target_category is None else np_sample[target_column] == target_category
	return np.average(imputed_values["p"], weights = weights) + np.average(original_values - imputed_values["np"], weights = imputed_weights)

def training_values(np_sample, p_sample, target_column, target_category = None, weights_column = None, covariates = None, psa_model = None, matching_model = None):
	training_weight = psa_weights(np_sample, p_sample, None, weights_column, covariates, psa_model)["np"]
	return matching_values(np_sample, p_sample, target_column, target_category, covariates, matching_model, training_weight)

def iqr(array):
	percentiles = np.nanpercentile(array, (25, 75), method = "median_unbiased")
	return percentiles[1] - percentiles[0]

def kw_weights(np_sample, p_sample, population_size = None, weights_column = None, covariates = None, model = None):
	np_size = np_sample.shape[0]
	my_propensities = propensities(np_sample, p_sample, weights_column, covariates, model)
	
	kernels = my_propensities[np_size:].reshape(-1, 1) - my_propensities[:np_size]
	m = min(np.std(kernels), iqr(kernels) / 1.349)
	h = 0.9 * m / pow(kernels.size, 1/5)
	kernels **= 2
	kernels /= -2*h*h
	np.exp(kernels, out = kernels)
	np.clip(kernels, 1.49012e-8, None, out = kernels)
	kernels /= np.sum(kernels, axis = 1).reshape(-1, 1)
	if weights_column is not None: kernels *= p_sample[weights_column].to_numpy().reshape(-1, 1)
	kernels = np.sum(kernels, axis = 0)
	if population_size: kernels *= population_size / np.sum(kernels)
	return kernels

def estimation(values, weights = None, axis = None):
	return np.average(values, weights = weights, axis = axis)

def confidence_interval(values, weights = None):
	data = (values,) if weights is None else (values, weights)
	confidence_interval = bootstrap(data, estimation, paired = True, n_resamples = 1000,
		random_state = 0, vectorized = True, batch = 1000).confidence_interval
	return (confidence_interval.low, confidence_interval.high)



import inps
import pandas as pd
import numpy as np
from numpy.random import default_rng
from xgboost import XGBRegressor, XGBClassifier
from sklearn.naive_bayes import BernoulliNB
from sklearn.neural_network import MLPRegressor

rng = default_rng(0)
pop_size = 10000
n = 1000
N = 2000
np_sample = rng.standard_normal(n * 5).reshape(-1, 5)
p_sample = rng.standard_normal(N * 3).reshape(-1, 3)
population = rng.standard_normal(pop_size * 3).reshape(-1, 3)
weights = [pop_size / N * 0.8] * int(N/2) + [pop_size / N * 1.2] * int(N/2)

def to_category(num_series):
	return pd.Series(np.where(num_series > 0, "Yes", "No"), dtype = "category", copy = False)

np_sample = pd.DataFrame(np_sample, columns = ["A", "B", "cat", "target", "target_cat"], copy = False)
p_sample = pd.DataFrame(p_sample, columns = ["A", "B", "cat"], copy = False)
population = pd.DataFrame(population, columns = ["A", "B", "cat"], copy = False)
np_sample["target_cat"] = to_category(np_sample["target_cat"])
np_sample["cat"] = to_category(np_sample["cat"])
p_sample["cat"] = to_category(p_sample["cat"])
population["cat"] = to_category(population["cat"])
p_sample["weights"] = weights

print(np_sample)
print(p_sample)

population_totals = pd.Series({"A": 10, "B": 5})
print(population_totals)
calibration_weights = inps.calibration_weights(np_sample, population_totals, population_size = pop_size)
print(calibration_weights)
calibration_weights2 = inps.calibration_weights(p_sample, population_totals, weights_column = "weights")
print(calibration_weights2)

mean_estimation = inps.estimation(np_sample["target"], calibration_weights)
mean_interval = inps.confidence_interval(np_sample["target"], calibration_weights)
print(mean_estimation)
print(mean_interval)
proportion_estimation = inps.estimation(np_sample["target_cat"] == "Yes", calibration_weights)
proportion_interval = inps.confidence_interval(np_sample["target_cat"] == "Yes", calibration_weights)
print(proportion_estimation)
print(proportion_interval)

psa_weights = inps.psa_weights(np_sample, p_sample, pop_size)
psa_weights = inps.psa_weights(np_sample, p_sample, pop_size, weights_column = "weights")
psa_weights = inps.psa_weights(np_sample, p_sample, pop_size, weights_column = "weights", covariates = ["A", "B", "cat"])
psa_weights2 = inps.psa_weights(np_sample, p_sample, pop_size, weights_column = "weights", model = XGBClassifier(enable_categorical = True, tree_method = "hist"))
psa_weights3 = inps.psa_weights(np_sample, p_sample, pop_size, weights_column = "weights", model = inps.make_preprocess_estimator(BernoulliNB()))

print(psa_weights["np"])
print(psa_weights["p"])

mean_estimation = inps.estimation(np_sample["target"], psa_weights["np"])
mean_interval = inps.confidence_interval(np_sample["target"], psa_weights["np"])
print(mean_estimation)
print(mean_interval)
proportion_estimation = inps.estimation(np_sample["target_cat"] == "Yes", psa_weights["np"])
proportion_interval = inps.confidence_interval(np_sample["target_cat"] == "Yes", psa_weights["np"])
print(proportion_estimation)
print(proportion_interval)

matching_values = inps.matching_values(np_sample, p_sample, "target")
cat_matching_values = inps.matching_values(np_sample, p_sample, "target_cat", "Yes")
matching_values = inps.matching_values(np_sample, p_sample, "target", covariates = ["A", "B", "cat"])
matching_values2 = inps.matching_values(np_sample, p_sample, "target", model = XGBRegressor(enable_categorical = True, tree_method = "hist"))
matching_values3 = inps.matching_values(np_sample, p_sample, "target", model = inps.make_preprocess_estimator(MLPRegressor()))

print(matching_values["p"])
print(matching_values["np"])
print(cat_matching_values["p"])
print(cat_matching_values["np"])

mean_estimation = inps.estimation(matching_values["p"], p_sample["weights"])
mean_interval = inps.confidence_interval(matching_values["p"], p_sample["weights"])
print(mean_estimation)
print(mean_interval)
proportion_estimation = inps.estimation(cat_matching_values["p"], p_sample["weights"])
proportion_estimation = inps.confidence_interval(cat_matching_values["p"], p_sample["weights"])
print(proportion_estimation)
print(proportion_interval)

doubly_robust_estimation = inps.doubly_robust_estimation(np_sample, p_sample, "target", covariates = ["A", "B", "cat"])
cat_doubly_robust_estimation = inps.doubly_robust_estimation(np_sample, p_sample, "target_cat", "Yes", covariates = ["A", "B", "cat"])
doubly_robust_estimation2 = inps.doubly_robust_estimation(np_sample, p_sample, "target", psa_model = XGBClassifier(enable_categorical = True, tree_method = "hist"), matching_model = XGBRegressor(enable_categorical = True, tree_method = "hist"))

print(doubly_robust_estimation)
print(cat_doubly_robust_estimation)

training_values = inps.training_values(np_sample, p_sample, "target", psa_model = XGBClassifier(enable_categorical = True, tree_method = "hist"), matching_model = XGBRegressor(enable_categorical = True, tree_method = "hist"))
print(training_values["p"])

kw_weights = inps.kw_weights(np_sample, p_sample, pop_size, weights_column = "weights", covariates = ["A", "B", "cat"])
print(kw_weights)

proportion_estimation = inps.estimation(np_sample["target_cat"] == "Yes", kw_weights)
print(proportion_estimation)

print(population)
imputed_values = inps.training_values(np_sample, population, "target")
cat_imputed_values = inps.training_values(np_sample, population, "target_cat", "Yes")
mean_estimation = inps.estimation(imputed_values["p"])
mean_interval = inps.confidence_interval(imputed_values["p"])
print(mean_estimation)
print(mean_interval)
proportion_estimation = inps.estimation(cat_imputed_values["p"])
proportion_interval = inps.confidence_interval(cat_imputed_values["p"])
print(proportion_estimation)
print(proportion_interval)
